"""NOTIFY message utilities."""
import struct
import asyncio
import socket
import logging
import xmltodict
import collections

from .error import PyMediaroomError

_LOGGER = logging.getLogger(__name__)

MEDIAROOM_BROADCAST_ADDR = "239.255.255.250"
MEDIAROOM_BROADCAST_PORT = 8082
TIMEOUT = 5
GEN_ID_FORMAT = "STB{}"

class MediaroomNotify(object):
    """Representation of the Mediaroom NOTIFY message."""
    def __init__(self, addr, data):
        self.src_ip = addr[0]
        self.src_port = addr[1]
        self._device = None
        while data[-1:] == '\0':
            data = data[:-1] # Strip trailing \0's
        while data[:6] != b"NOTIFY":
            data = data[1:] # Strip head garbage
        self.data = data

        for line in data.decode().split('\n'):
            if line.startswith("x-type"):
                self._type = line[line.find(":")+2:]
            elif line.startswith("x-filter"):
                self._filter = line[line.find(":")+2:]
            elif line.startswith("x-lastUserActivity"):
                self._last_user_activity = line[line.find(":")+2:]
            elif line.startswith("x-device"):
                self._device = line[line.find(":")+2:]
            elif line.startswith("x-debug") or\
                line.startswith("x-location") or\
                line.startswith("NOTIFY") or\
                line.startswith("\r"):
                pass
            elif line.startswith("<"):
                self._node = xmltodict.parse(line)['node']
#                import pprint
#                pprint.pprint(self._node)
            else:
                _LOGGER.error("UNKNOWN LINE: %s", line)

    def __str__(self):
#        _LOGGER.debug("x-type: %s", self._type)
#        _LOGGER.debug("x-filter: %s", self._filter)
#        _LOGGER.debug("x-lastUserActivity: %s", self._last_user_activity)
#        _LOGGER.debug("x-device: %s", self._device)

        return "NOTIFY from {} - {}".format(self.src_ip, self.tune)

    @property
    def tune(self):
        """XML node representing tune."""
        if self._node.get('activities'):
            tune = self._node['activities'].get('tune')
            if type(tune) is collections.OrderedDict:
                return tune
            elif type(tune) is list:
                return tune[0]
            return tune
        return None

    @property
    def stopped(self):
        """Return if the stream is stopped."""
        if self.tune and self.tune.get('@stopped'):
            return True if self.tune.get('@stopped') == 'true' else False
        else:
            raise PyMediaroomError("No information in <node> about @stopped")

    @property
    def timeshift(self):
        """Return if the stream is a timeshift."""
        if self.tune and self.tune.get('@src'):
            return True if self.tune.get('@src').startswith('timeshift') else False
        else:
            raise PyMediaroomError("No information in <node> about @src")

    @property
    def recorded(self):
        """Return if the stream is a recording."""
        if self.tune and self.tune.get('@src'):
            return True if self.tune.get('@src').startswith('mbr') else False
        else:
            raise PyMediaroomError("No information in <node> about @src")

    @property
    def ip_address(self):
        """Return IP address of the STB."""
        return self.src_ip

    @property
    def device_uuid(self):
        """Return device UUID."""
        if self._device:
            return self._device
        return GEN_ID_FORMAT.format(self.src_ip)

def create_socket():
    # Create multicast socket
    addrinfo = socket.getaddrinfo(MEDIAROOM_BROADCAST_ADDR, None)[0]
    sock = socket.socket(addrinfo[0], socket.SOCK_DGRAM)
    sock.settimeout(0)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # for BSD/Darwin only
    try:
        socket.SO_REUSEPORT
    except AttributeError:
        _LOGGER.debug("No SO_REUSEPORT available, skipping")
    else:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

    # IGMP packet
    group_bin = socket.inet_pton(addrinfo[0], addrinfo[4][0])
    mreq = group_bin + struct.pack('=I', socket.INADDR_ANY)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

    sock.bind(('', MEDIAROOM_BROADCAST_PORT))

    return sock

class MediaroomProtocol(asyncio.DatagramProtocol):
    """Mediaroom asyncio protocol."""

    def __init__(self, responses_callback, box_ip=None):
        super().__init__()
        self.responses = responses_callback
        self.box_ip = box_ip

    def connection_made(self, transport):
        """Setup transport."""
        self.transport = transport

    def connection_lost(self, exception):
        """Connection is lost."""
        if exception is None:
            pass
        else:
            _LOGGER.error("Received unexpected connection loss: %s", exception)

    def datagram_received(self, data, addr):
        """Datagram received callback."""
        #_LOGGER.debug(data)
        if not self.box_ip or self.box_ip == addr[0]:
            self.responses(MediaroomNotify(addr, data))

    def error_received(self, exception):
        """Datagram error callback."""
        if exception is None:
            pass
        else:
            import pprint
            pprint.pprint(exception)
            _LOGGER.error('Error received: %s', exception)

    def close(self):
        """Close socket."""
        _LOGGER.debug("Closing MediaroomProtocol")
        self.transport.close()

async def install_mediaroom_protocol(responses_callback, box_ip=None):
    """Install an asyncio protocol to process NOTIFY messages."""
    from . import version
    _LOGGER.debug(version)

    loop = asyncio.get_event_loop()

    mediaroom_protocol = MediaroomProtocol(responses_callback, box_ip)

    sock = create_socket()

    await loop.create_datagram_endpoint(lambda: mediaroom_protocol, sock=sock)

    return mediaroom_protocol



"""Remote control agent for Mediaroom Set up Boxes (STB)."""
import logging
import asyncio
from enum import Enum
from async_timeout import timeout

from .commands import COMMANDS
from .notify import (install_mediaroom_protocol, GEN_ID_FORMAT)
from .error import PyMediaroomError

_LOGGER = logging.getLogger(__name__)

REMOTE_CONTROL_PORT = 8082

GET_STATE_TIMEOUT = 10
OPEN_CONTROL_TIMEOUT = 5

class State(Enum):
    """Available States."""
    OFF = 0
    STANDBY = 1 
    PLAYING_LIVE_TV = 2
    PLAYING_RECORDED_TV = 3
    PLAYING_TIMESHIFT_TV = 4
    STOPPED = 5
    UNKNOWN = -1

class Remote():
    """This class represents a MediaRoom Set-up-Box Remote Control."""

    def __init__(self, ip):
        self.stb_ip = ip
        self._state = State.UNKNOWN
        self.lock = asyncio.Lock()
        self.tune_src = None
        self.current_channel = None

    @property
    def device_id(self):
        """Generate pseudo device ID from fixed IP assigned by operator."""
        return GEN_ID_FORMAT.format(self.stb_ip)

    @property
    def state(self):
        return self._state

    def resolv(self, tune_src):
        """Return the channel name. TODO: implement remote webservice to resolv tune_src"""
        self.tune_src = tune_src
        return self.current_channel

    def __repr__(self):
        return self.stb_ip

    async def send_cmd(self, cmd):
        """Send remote command to STB."""
        _LOGGER.info("Send cmd = %s", cmd)

        if cmd not in COMMANDS and cmd not in range(0, 999):
            _LOGGER.error("Unknown command")
            raise PyMediaroomError("Unknown commands")

        keys = []
        if cmd in range(0, 999):
            for character in str(cmd):
                keys.append(COMMANDS["Number"+str(character)])
            if len(keys) < 3:
                keys.append(COMMANDS["OK"])
            self.current_channel = cmd
        else:
            keys = [COMMANDS[cmd]]

        try:
            async with timeout(OPEN_CONTROL_TIMEOUT):
                async with self.lock:
                    _LOGGER.debug("Connecting to %s:%s", self.stb_ip, REMOTE_CONTROL_PORT)
                    stb_recv, stb_send = await asyncio.open_connection(self.stb_ip,
                                                                       REMOTE_CONTROL_PORT,
                                                                       )
                    await stb_recv.read(6)
                    _LOGGER.info("Connected to %s:%s", self.stb_ip, REMOTE_CONTROL_PORT)

                    for key in keys:
                        _LOGGER.debug("%s key=%s", cmd, key)
                        stb_send.write("key={}\n".format(key).encode('UTF-8'))
                        _ = await stb_recv.read(3)
                        await asyncio.sleep(0.300)

        except asyncio.TimeoutError as error:
            _LOGGER.warning(error)
            raise PyMediaroomError("Timeout connecting to {}".format(self.stb_ip))
        except ConnectionRefusedError as error:
            _LOGGER.warning(error)
            raise PyMediaroomError("Connection refused to {}".format(self.stb_ip))

    async def turn_on(self):
        """Turn off media player."""
        _LOGGER.debug("turn_on while %s", self._state)
        if self._state in [State.STANDBY, State.OFF, State.UNKNOWN]:
            await self.send_cmd('Power')
            self._state = State.PLAYING_LIVE_TV
        return self._state

    async def turn_off(self):
        """Turn off media player."""
        _LOGGER.debug("turn_off while %s", self._state)
        if self._state in [State.PLAYING_LIVE_TV, State.PLAYING_RECORDED_TV,
                          State.PLAYING_TIMESHIFT_TV, State.UNKNOWN]:
            await self.send_cmd('Power')
            self._state = State.OFF
        return self._state

    def notify_callback(self, notify):
        """Process State from NOTIFY message."""
        _LOGGER.debug(notify)
        if notify.ip_address != self.stb_ip:
            return

        if notify.tune:
            self._state = State.PLAYING_LIVE_TV 
            self.tune_src = notify.tune['@src']
            try:
                if notify.stopped:
                    self._state = State.STOPPED
                elif notify.timeshift:
                    self._state = State.PLAYING_TIMESHIFT_TV
                elif notify.recorded:
                    self._state = State.PLAYING_RECORDED_TV
            except PyMediaroomError as e:
                _LOGGER.debug("%s please report at https://github.com/dgomes/pymediaroom/issues", e)
        else:
            self._state = State.STANDBY
        
        _LOGGER.debug("%s is %s", self.stb_ip, self._state)
        return self._state

async def discover(ignore_list=[], max_wait=30):
    """List STB in the network."""
    stbs = []
    try:
        async with timeout(max_wait):
            def responses_callback(notify):
                """Queue notify messages."""
                _LOGGER.debug("Found: %s", notify.ip_address)
                stbs.append(notify.ip_address)

            mr_protocol = await install_mediaroom_protocol(responses_callback=responses_callback)
            await asyncio.sleep(max_wait)
    except asyncio.TimeoutError:
        mr_protocol.close()
        _LOGGER.debug("discover() timeout!")

    return list(set([stb for stb in stbs if stb not in ignore_list]))



from setuptools import setup

version=open("VERSION").read()
long_description=open("README.rst").read()

setup(name='pymediaroom',
      version=version,
      description='Remote control your Mediaroom Set-up-box',
      long_description=long_description,
      long_description_content_type='text/x-rst',
      url='http://github.com/dgomes/pymediaroom',
      author='Diogo Gomes',
      author_email='diogogomes@gmail.com',
      license='MIT',
      packages=['pymediaroom'],
      install_requires=[
        'async_timeout>=4.0',
        'xmltodict',
      ],
      python_requires='>=3.5',
      download_url= 'https://github.com/dgomes/pymediaroom/tarball/0.6.5.2',
      )
